/***************************************************************************
                          ann.h  -  description
                             -------------------
    begin                : Sat Dec 1 2001
    copyright            : (C) 2001 by Gene Ruebsamen
    email                : gene@erachampion.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef ANN_H
#define ANN_H
#include <iostream>
#include "net.h"

using std::ostream;

class ANN {
	friend ostream &operator<<(ostream &, const ANN &);		// overloaded << operator
public:
	ANN(int=11, int=8, int=6, int=8);		// #inputs, #hidden neurons, #outputs, #recurrent layers
	ANN(const ANN &);			// copy constructor
	~ANN();								// destructor

	bool operator==(const ANN &) const;											// equality
	bool operator!=(const ANN &right) const
		{ return !(*this == right); }
	const ANN *operator+(const ANN &right);		// reproduction
	const ANN &operator=(const ANN &right);		// assignment operator

	void resetANN();											// initialize network with random weights
	const float *getWeights() const;			// get a vector of the weights
	float getFitness(void);									// get fitness of the current ANN
	void setFitness(float);							// set fitness of individual
	int getLength() const;								// get chromosome (weight) length
 	const float * const getOutput(const float * const);			// pass in input, and get output
	float *out;
private:
	void setWeights(float *weights);
	Net neuralNet;
	float fit_score;
	float *weightPtr;
};


#endif

