/***************************************************************************
                          biped.cpp  -  Main Program
                             -------------------
    begin                : Sat Dec 1 2001
    copyright            : (C) 2001 by Gene Ruebsamen
    email                : gene@erachampion.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <ode/ode.h>
#include <drawstuff/drawstuff.h>
#include "net.h"
#include "ann.h"
#include "creature.h"
#include "runner.h"
#include "constants.h"
#include <iostream>
#include <cstdlib>
#include <ctime>
#include <iomanip>
#include <fstream>

using std::cout;
using std::endl;
using std::setw;
using std::ios;
using std::cerr;
using std::ofstream;

#ifdef MSVC
#pragma warning(disable:4244 4305)  // for VC++, no precision loss complaints
#endif

// select correct drawing functions

#ifdef dDOUBLE
#define dsDrawBox dsDrawBoxD
#define dsDrawSphere dsDrawSphereD
#define dsDrawCylinder dsDrawCylinderD
#define dsDrawCappedCylinder dsDrawCappedCylinderD
#endif

// constants
bool all = true;
int display=0;
int maxtime = MAXTIME;
// environment setup
static dWorldID earth;
static dGeomID ground[NUMHUMANS];
static dJointGroupID contactgroup;
static dSpaceID environment[NUMHUMANS];

// add additional obstacles
/*  static dGeomID ground_box[NUMHUMANS];
    static dGeomID ground_box2[NUMHUMANS];
    static dGeonID ground_box3[NUMHUMANS];
*/

// GA Setup
ANN *pop;
ANN *newpop;
bool foot1=false, foot2=false;
float fitarray[POPSIZE], total;
int curhuman=0, generation=0;

// output file setup
ofstream outFile("output.dat",ios::out);

// global creatures
Runner *human[NUMHUMANS];		// array of pointers

// function prototypes
void updateGA();
int getParent(float);

// Collision call back function here
static void nearCallback (void *data, dGeomID o1, dGeomID o2)
{
	int i,n;

	// only collide things with the ground
	dBodyID b1 = dGeomGetBody(o1);
	dBodyID b2 = dGeomGetBody(o2);
	if (b1 && b2 && dAreConnected (b1,b2)) return;

	const int N = 10;	// approximate 10 contacts
	dContact contact[N];
	n = dCollide (o1,o2,N,&contact[0].geom,sizeof(dContact));

  // contact! 
	if (n > 0) {
 		human[curhuman]->footContact(o1,o2,ground[curhuman]);
 		human[curhuman]->bodyContact(o1,o2,ground[curhuman]);
		for (i=0; i<n; i++) {
			if ((ground[curhuman] == o1) || (ground[curhuman] == o2)) {
				contact[i].surface.mode = dContactSoftERP | dContactSoftCFM | dContactSlip1 | dContactSlip2;
				contact[i].surface.mu = dInfinity;
				contact[i].surface.slip1 = 0.2; // lower is more friction 0.8
				contact[i].surface.slip2 = 0.2;
				contact[i].surface.soft_erp = 0.2;
				contact[i].surface.soft_cfm = 0.05;
			} else {
				contact[i].surface.mode = dContactSoftERP | dContactSoftCFM | dContactSlip1 | dContactSlip2;
				contact[i].surface.mu = dInfinity;
				contact[i].surface.slip1 = 0.1; // lower is more friction 0.4
				contact[i].surface.slip2 = 0.1;
				contact[i].surface.soft_erp = 0.2;
				contact[i].surface.soft_cfm = 0.05;
			}
			dJointID c = dJointCreateContact (earth,contactgroup,&contact[i]);
			dJointAttach (c, dGeomGetBody(contact[i].geom.g1), dGeomGetBody(contact[i].geom.g2));
		}
	}
}

//----------
// start simulation - set viewpoint
//----------
static void start()
{
	static float xyz[3] = {0.8317f,-0.9817f,1.8000f};
  static float hpr[3] = {121.0000f,-27.5000f,0.0000f};
  dsSetViewpoint (xyz,hpr);
  printf ("Press:\t'a' to increase speed.\n"
	  "\t'z' to decrease speed.\n"
	  "\t',' to steer left.\n"
	  "\t'.' to steer right.\n"
	  "\t' ' to reset speed and steering.\n");
}

//----------
// Respond to Keypresses during simulation
//----------
static void command (int cmd)
{
  switch (cmd) {
  case 'a': case 'A':
		all = true;
	break;
  case 'z': case 'Z':
		all = false;
  break;
  case 'n': case 'N' :
		if (display < NUMHUMANS-1)
			display++;
		else
			display=0;
	break;
  case 'p': case 'P' :
		if (display > 0)
			display--;
		else
			display=NUMHUMANS-1;
	break;
	case 'i': case 'I' :
  maxtime +=50;
  cout << "MaxTime: " << maxtime << endl;
	break;
	case 'd': case 'D' :
  maxtime -=50;
  cout << "MaxTime: " << maxtime << endl;
	break;

  case 'm': case 'M' :
    human[display]->printStats();
  break;
  }
}

// run Simulation
static void simLoop(int pause)
{
	float ann_input[NUMHUMANS][24];	// 18 joint position inputs (0-17) + 3 position elevation (18) + 3 rotation (19,20,21)
	static int curtime=0;
  static int count=0;
	bool done=false;

  dVector3 ss;

  // clear input Sensors
  for (int j=0;j<NUMHUMANS;j++)
    for (int k=0;k<24;k++)
      ann_input[j][k] = 0;

	dsSetColor (0.7,0.3,0.7);

// Use the following to add obstacles to the simulation
/*  dGeomBoxGetLengths (ground_box[0],ss);
    dsDrawBox (dGeomGetPosition(ground_box[0]),dGeomGetRotation(ground_box[0]),ss);

    dGeomBoxGetLengths (ground_box2[0],ss);
    dsDrawBox (dGeomGetPosition(ground_box2[0]),dGeomGetRotation(ground_box2[0]),ss);
*/

	if (!pause)
	{
		for (curhuman=0;curhuman<NUMHUMANS;curhuman++)
    {
      human[curhuman]->clearSensor();
			dSpaceCollide (environment[curhuman],0,&nearCallback);
    }
		// input to network
		if (count >= ANNSTEP)
		{
			done = true;
			for (curhuman=0;curhuman<NUMHUMANS;curhuman++)
				if (!human[curhuman]->hasFallen()) 
					human[curhuman]->getInputs(ann_input[curhuman]);		// get input values for Neural Network
		}
		dWorldStep (earth,SIMSTEP);          			// step simulation by one
		dJointGroupEmpty (contactgroup);					// remove all contact joints

		if (count >= ANNSTEP)
			for (curhuman=0;curhuman<NUMHUMANS;curhuman++)
			{                    		 // fix this problem ->|<- pop + curhuman can go over POPSIZE
				if (!human[curhuman]->hasFallen()) {
					done = false;
					human[curhuman]->setMuscles(pop[curhuman].getOutput(ann_input[curhuman]));    // let the brain figure out what to set muscles to
				}  else {
					human[curhuman]->shutoff();
				}
			}
		curtime++;
	}		// end if (!pause)
	if (count >= ANNSTEP)
		count = 0;
	else
		count++;
	if (all)
	{
		for (int j=0;j<NUMHUMANS;j++)
		{
			if (j == display)
  			human[j]->draw(true);
			else
	 			human[j]->draw();  
		}
	}
	else
	{
		human[display]->draw();
	}

	if (curtime >= maxtime || (done == true))
	{
		for (int k=0;k<NUMHUMANS;k++)
		{
			fitarray[k] = human[k]->getFitness();	// get fitness of individual

			delete human[k];
 	 	  human[k] = new Runner(earth,environment[k],(k*1.5)-NUMHUMANS/2,0,STARTZ);
		}
  	updateGA();		// update population (population++)
		// reset
		curtime = 0;
	}	
}	// end simLoop Function

//--------
// Update the Genetic Algorithm
//--------
void updateGA()
{
  float avg=0, min=1000, max=0, max2=0,max3=0,max4=0;
	int maxindex,max2index,max3index,max4index;


	// calculate statistics
  for (int i=0;i<POPSIZE;i++)
	{
		avg += fitarray[i];
		if (fitarray[i] > max)
		{
			max = fitarray[i];
			maxindex=i;
		}
    else if (fitarray[i] > max2)
    {
      max2 = fitarray[i];
      max2index=i;
    }
    else if (fitarray[i] > max3)
    {
      max3 = fitarray[i];
      max3index=i;
    }
    else if (fitarray[i] > max4)
    {
      max4 = fitarray[i];
      max4index=i;
    }
		else if (fitarray[i] < min)
			min = fitarray[i];			
	}

	const ANN *tempANN;
	int index1, index2;
	total = avg;
	avg /= (float)POPSIZE;
	cout << "GEN: " << setw(3) << generation << ", AVG: " << setw(7) << avg << ", MAX: " << max << " " << max2 << " " << max3 << ", MIN: " << min << endl;
  outFile << generation << "," << avg << "," << max << endl;

	// linear fitness scaling routines
/*		const double fmultiple=2.0;		
		double delta,a,b, sumfitness, fitness;
		sumfitness=0.0;
		if (min > ((fmultiple*avg - max) / (fmultiple - 1.0))) {
			delta = max - avg;
			a = (fmultiple - 1.0) * avg/delta;
			b = avg * (max - fmultiple*avg) / delta;
		} else {
			delta = avg - min;
			a = avg / delta;
			b = -min * avg / delta;
		}
		for (int j=0;j<POPSIZE;j++)
		{
			fitness = a * fitarray[j] + b;
			fitarray[j]=fitness;
			sumfitness += fitness;	
		}	      */

	// let the best guy live in the next gen
	newpop[POPSIZE-1] = pop[maxindex];
  newpop[POPSIZE-2] = pop[max2index];
  newpop[POPSIZE-3] = pop[max3index];
  newpop[POPSIZE-4] = pop[max4index];
	
	// roulette wheel selection		
	for (int i=0;i<POPSIZE-4;i++)	// for best guy live on
//		for (int i=0;i<POPSIZE;i++)		// for fitness scaling
	{
		index1 = getParent(total);		
    index2 = getParent(total);
		tempANN = (pop[index1] + pop[index2]);			// mating
		newpop[i] = *tempANN;
		delete tempANN;
	}				
		// copy new pop
		for (int j=0;j<POPSIZE;j++)
			pop[j] = newpop[j];

	generation++;
	for (int j=0;j<POPSIZE;j++)
		fitarray[j]=0;
}

int getParent(float total)
{
		double roulette, sum=0;
		int count=0;
		roulette = (((float)(rand() %1000))/1000.0)*total;
		
		while (sum < roulette)
			sum += fitarray[count++];	
		if (count > 0)
			count--;
		return count;
}	


// main program
int main(int argc, char *argv[]) {
	// setup pointers to drawstuff callback functions
	dsFunctions fn;
	fn.version = DS_VERSION;
	fn.start = &start;
	fn.step = &simLoop;
	fn.command = &command;
	fn.stop = 0;
	fn.path_to_textures = "/home/gene/thesis/textures";
	srand(time(0));			// seed random number generator

  if (!outFile) {
    cerr << "File could not be opened" << endl;
    exit(1);
  }

  outFile << "Generation, avg, max" << endl;
	
	// create world (setup environment)
	earth = dWorldCreate();		// create ODE world

  dMatrix3 R,R2;
  dRFromAxisAndAngle (R,1,0,0,-0.10);
  dRFromAxisAndAngle (R2,1,0,0,0.10);


	for (int i=0;i<NUMHUMANS;i++) {
		environment[i] = dHashSpaceCreate();
// Add environmental obstacles
/*    ground_box[i] = dCreateBox (environment[i],160,5.5,2);
      ground_box2[i] = dCreateBox (environment[i],160,5.5,2);
      dGeomSetPosition (ground_box[i],0,-5,-0.75);
      dGeomSetPosition (ground_box2[i],0,-10,-0.75);
      dGeomSetRotation (ground_box[i],R);
      dGeomSetRotation (ground_box2[i],R2);
*/
  }


  contactgroup = dJointGroupCreate (0);
	dWorldSetGravity (earth,0,0,-GRAVITY);	// set gravity (x,y,z)
	dWorldSetERP (earth,ERP);			// global ERP
	dWorldSetCFM (earth,CFM);			// global CFM
	pop = new ANN[POPSIZE];
	newpop = new ANN[POPSIZE];

	// create creatures forward
	for (int i=0;i<NUMHUMANS;i++)	{
      ground[i] = dCreatePlane(environment[i],0,0,1,0);
  	 	human[i] = new Runner(earth,environment[i],(i*1.5)-NUMHUMANS/2,0,STARTZ);
	}

	// Start Simulation
	dsSimulationLoop (argc,argv,640,480,&fn);	// 640 x 480 OpenGL display

	// Destroy World (end program)	
	for (int i=0;i<NUMHUMANS;i++) {
	  delete human[i];
		dSpaceDestroy(environment[i]);
	}
	dJointGroupDestroy (contactgroup);
	dWorldDestroy(earth);

	delete [] pop;
	delete [] newpop;

	return 0;	
}
