/***************************************************************************
                          box.cpp  -  description
                             -------------------
    begin                : Thu Feb 21 2002
    copyright            : (C) 2002 by Gene Ruebsamen
    email                : 
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


#include "box.h"      // import constants for bodies
#include <iostream>

using std::cout;
using std::endl;

Box::Box(dWorldID world, dSpaceID space, dReal x, dReal y, dReal z, bool reverse) :
	Creature(world,space,x,y,z), numBody(7), numGeom(7), numJoint(6)		// Build Creature
{
	// Body Setup
	TORSO[0] = 0.7; TORSO[1] = 1.2; TORSO[2] = 0.5;
	LEG1[0] = 0.05; LEG1[1] = 0.05; LEG1[2] = 0.67;
	LEG2[0] = 0.05; LEG2[1] = 0.05; LEG2[2] = 0.67;
	LEG3[0] = 0.35; LEG3[1] = 0.05; LEG3[2] = 0.47;
	GAP = 0.01;

	l_foot_sens = false;					// init foot sensors
	r_foot_sens = false;
	body = new dBodyID[numBody];	// allocate space for Bodies
	geom = new dGeomID[numGeom];	// allocate space for Geometries	
	joint = new dJointID[numJoint];	// allocate space for Joints
	
	// create new Joint Group
	jGroup = dJointGroupCreate(0);
	//geomGroup = dCreateGeomGroup(space);   // dont use if not needed (slowdown)

	// constants (Mass)
	const dReal MTORSO=2.8;
	const dReal MULEG=0.05;
	const dReal MMLEG=0.05;
	const dReal MLLEG=0.08;

	dMass m;
  dMatrix3 R;

	// create bodies
	for (int i=0;i<numBody;i++)
		body[i]=dBodyCreate(world);
  	dBodySetPosition (body[0],0+x,0+y,z);	// torso
  	dBodySetPosition (body[1],TORSO[0]/3+x,y-TORSO[1]/5,z+(TORSO[2]/2+LEG1[2]/2+GAP));	//right upper leg
   	dBodySetPosition (body[2],-TORSO[0]/3+x,y-TORSO[1]/5,z+(TORSO[2]/2+LEG1[2]/2+GAP));	// left upper leg
  	dBodySetPosition (body[3],TORSO[0]/3+GAP/2+x,y-TORSO[1]/5,z+TORSO[2]/2+LEG1[2]+LEG2[2]/2+GAP);	//right foot
  	dBodySetPosition (body[4],-TORSO[0]/3-GAP/2+x,y-TORSO[1]/5,z+TORSO[2]/2+LEG1[2]+LEG2[2]/2+GAP);	//left foot
  	dBodySetPosition (body[5],TORSO[0]/3+GAP/2+x,y-TORSO[1]/5,z+TORSO[2]/2+LEG1[2]+LEG2[2]+LEG3[2]/2+2*GAP);	//right foot
  	dBodySetPosition (body[6],-TORSO[0]/3-GAP/2+x,y-TORSO[1]/5,z+TORSO[2]/2+LEG1[2]+LEG2[2]+LEG3[2]/2+2*GAP);	//left foot

	// torso
	dMassSetBox (&m,1,TORSO[0],TORSO[1],TORSO[2]);	// torso
	dMassAdjust (&m,MTORSO);	// torso weight
	dBodySetMass (body[0],&m);
	geom[0] = dCreateBox (space,TORSO[0],TORSO[1],TORSO[2]);
	dGeomSetBody (geom[0],body[0]);

	// upper legs
	dMassSetBox (&m,1,LEG1[0],LEG1[1],LEG1[2]);	// legs
	dMassAdjust (&m,MULEG);	// upper leg weight
	for (int i=1;i<=2;i++)
	{
		dBodySetMass (body[i],&m);
		geom[i] = dCreateBox (space,LEG1[0],LEG1[1],LEG1[2]);
		dGeomSetBody (geom[i],body[i]);
	}

	// middle legs
	dMassSetBox (&m,1,LEG2[0],LEG2[1],LEG2[2]);	// legs
	dMassAdjust (&m,MMLEG);	// middle leg weight
	for (int i=3;i<=4;i++)
	{
		dBodySetMass (body[i],&m);
		geom[i] = dCreateBox (space,LEG1[0],LEG1[1],LEG1[2]);
		dGeomSetBody (geom[i],body[i]);
	}

	// lower legs
	dMassSetBox (&m,1,LEG3[0],LEG3[1],LEG3[2]);	// legs
	dMassAdjust (&m,MLLEG);	// middle leg weight
	for (int i=5;i<=6;i++)
	{
		dBodySetMass (body[i],&m);
		geom[i] = dCreateBox (space,LEG1[0],LEG1[1],LEG1[2]);
		dGeomSetBody (geom[i],body[i]);
	}

	/*for (int i=0;i<5;i++)
		dGeomGroupAdd(geomGroup,geom[i]); */

	/********************************** CREATE JOINTS **********************************/
	// right Leg Socket
	joint[0] = dJointCreateHinge(world,jGroup);
	dJointAttach(joint[0],body[0],body[1]);
	dJointSetHingeAnchor(joint[0],TORSO[0]/3+x,y-TORSO[1]/5,z+(TORSO[2]/2+GAP/2));
	dJointSetHingeParam(joint[0],dParamLoStop,-1.1);
	dJointSetHingeParam(joint[0],dParamHiStop,0);
	dJointSetHingeParam(joint[0],dParamStopERP,0.2);
	dJointSetHingeParam(joint[0],dParamStopCFM,0.05);
	dJointSetHingeParam(joint[0],dParamBounce,0.3);
	dJointSetHingeParam(joint[0],dParamFudgeFactor,0.5);
	dJointSetHingeParam(joint[0],dParamVel,0);
	dJointSetHingeParam(joint[0],dParamFMax,0.5);

	// left Leg Socket
	joint[1] = dJointCreateHinge(world,jGroup);
	dJointAttach(joint[1],body[0],body[2]);
	dJointSetHingeAnchor(joint[1],-TORSO[0]/3+x,y-TORSO[1]/5,z+(TORSO[2]/2+GAP/2));
	dJointSetHingeParam(joint[1],dParamLoStop,-1.1);
	dJointSetHingeParam(joint[1],dParamHiStop,0);
	dJointSetHingeParam(joint[1],dParamStopERP,0.2);
	dJointSetHingeParam(joint[1],dParamStopCFM,0.05);
	dJointSetHingeParam(joint[1],dParamBounce,0.3);
	dJointSetHingeParam(joint[1],dParamFudgeFactor,0.5);
	dJointSetHingeParam(joint[1],dParamVel,0);
	dJointSetHingeParam(joint[1],dParamFMax,0.5);


	// right middle leg
	joint[2] = dJointCreateHinge(world,jGroup);
	dJointAttach(joint[2],body[1],body[3]);
	dJointSetHingeAnchor(joint[2],TORSO[0]/3+GAP/2+x,y-TORSO[1]/5,z+TORSO[2]/2+LEG1[2]+GAP);
	dJointSetHingeAxis(joint[2],1,0,0);
	dJointSetHingeParam(joint[2],dParamLoStop,-1.5);
	dJointSetHingeParam(joint[2],dParamHiStop,0);
	dJointSetHingeParam(joint[2],dParamStopERP,0.3);
	dJointSetHingeParam(joint[2],dParamStopCFM,0.05);
	dJointSetHingeParam(joint[2],dParamBounce,0.3);
	dJointSetHingeParam(joint[2],dParamFudgeFactor,0.5);
	dJointSetHingeParam(joint[2],dParamVel,0);
	dJointSetHingeParam(joint[2],dParamFMax,0.5);
	// left middle foot
	joint[3] = dJointCreateHinge(world,jGroup);
	dJointAttach(joint[3],body[2],body[4]);
	dJointSetHingeAnchor(joint[3],-TORSO[0]/3+GAP/2+x,y-TORSO[1]/5,z+TORSO[2]/2+LEG1[2]+GAP);
	dJointSetHingeAxis(joint[3],1,0,0);
	dJointSetHingeParam(joint[3],dParamLoStop,-1.5);
	dJointSetHingeParam(joint[3],dParamHiStop,0.0);
	dJointSetHingeParam(joint[3],dParamStopERP,0.3);
	dJointSetHingeParam(joint[3],dParamStopCFM,0.05);
	dJointSetHingeParam(joint[3],dParamBounce,0.3);
	dJointSetHingeParam(joint[3],dParamFudgeFactor,0.5);
	dJointSetHingeParam(joint[3],dParamVel,0);
	dJointSetHingeParam(joint[3],dParamFMax,0.5);
	// right lower leg
	joint[4] = dJointCreateHinge(world,jGroup);
	dJointAttach(joint[4],body[3],body[5]);
	dJointSetHingeAnchor(joint[4],TORSO[0]/3+GAP/2+x,y-TORSO[1]/5,z+TORSO[2]/2+LEG1[2]+LEG2[2]+(3*GAP)/2);
	dJointSetHingeAxis(joint[4],1,0,0);
	dJointSetHingeParam(joint[4],dParamLoStop,-1.4);
	dJointSetHingeParam(joint[4],dParamHiStop,0.1);
	dJointSetHingeParam(joint[4],dParamStopERP,0.5);
	dJointSetHingeParam(joint[4],dParamStopCFM,0.07);
	dJointSetHingeParam(joint[4],dParamBounce,0.2);
	dJointSetHingeParam(joint[4],dParamFudgeFactor,0.5);
	dJointSetHingeParam(joint[4],dParamVel,0);
	dJointSetHingeParam(joint[4],dParamFMax,0.5);
	// left lower leg
	joint[5] = dJointCreateHinge(world,jGroup);
	dJointAttach(joint[5],body[4],body[6]);
	dJointSetHingeAnchor(joint[5],-TORSO[0]/3+GAP/2+x,y-TORSO[1]/5,z+TORSO[2]/2+LEG1[2]+LEG2[2]+(3*GAP)/2);
	dJointSetHingeAxis(joint[5],1,0,0);
	dJointSetHingeParam(joint[5],dParamLoStop,-1.4);
	dJointSetHingeParam(joint[5],dParamHiStop,0.1);
	dJointSetHingeParam(joint[5],dParamStopERP,0.5);
	dJointSetHingeParam(joint[5],dParamStopCFM,0.07);
	dJointSetHingeParam(joint[5],dParamBounce,0.2);
	dJointSetHingeParam(joint[5],dParamFudgeFactor,0.5);
	dJointSetHingeParam(joint[5],dParamVel,0);
	dJointSetHingeParam(joint[5],dParamFMax,0.5);

  if (!reverse) {
    dRFromAxisAndAngle (R, 0,0,0,0);
    for (int i=0;i<7;i++)
      dBodySetRotation(body[i], R);
  } else {
    dRFromAxisAndAngle (R, 0,0,1,3.1416);  // rotate the creatures
    for (int i=0;i<7;i++)
      dBodySetRotation(body[i], R);
  }


}

Box::~Box()
{
	// destroy joint group
	dJointGroupDestroy(jGroup);
	for (int i=0;i<numGeom;i++) {
//		dGeomGroupRemove(geomGroup,geom[i]);
		dGeomDestroy(geom[i]);
	}
	// destroy bodies
	for (int i=0;i<numBody;i++)
		dBodyDestroy(body[i]);

	// deallocate space
	delete [] body;
	delete [] geom;
	delete [] joint;
}


void Box::draw(bool selected) const
{
	dsSetColor (0,1,1);
	dsSetTexture (DS_WOOD); // DS_WOOD
	// head
	dsDrawBox(dBodyGetPosition(body[0]),dBodyGetRotation(body[0]),TORSO);
	if (selected)
		dsSetColor (1,0,0);
	else
		dsSetColor (1,0,1);
	// legs
	dsDrawBox(dBodyGetPosition(body[1]),dBodyGetRotation(body[1]),LEG1);
	dsDrawBox(dBodyGetPosition(body[2]),dBodyGetRotation(body[2]),LEG1);
	dsDrawBox(dBodyGetPosition(body[3]),dBodyGetRotation(body[3]),LEG2);
	dsDrawBox(dBodyGetPosition(body[4]),dBodyGetRotation(body[4]),LEG2);
	dsDrawBox(dBodyGetPosition(body[5]),dBodyGetRotation(body[5]),LEG3);
	dsDrawBox(dBodyGetPosition(body[6]),dBodyGetRotation(body[6]),LEG3);
}

void Box::setMuscles(const float * const inputs)
{
	// right Leg Socket
//	dJointSetAMotorParam(joint[6],dParamVel,(dReal) inputs[16]*1.55);
//	dJointSetAMotorParam(joint[6],dParamFMax, (dReal)fabs(inputs[17]*1.95));
	dJointSetHingeParam(joint[0],dParamVel,(dReal) inputs[0]*2.55);
	dJointSetHingeParam(joint[0],dParamFMax,(dReal)2.55);
	// left leg socket
//	dJointSetAMotorParam(joint[7],dParamVel,(dReal) inputs[22]*1.55);
//	dJointSetAMotorParam(joint[7],dParamFMax,(dReal)fabs(inputs[23]*1.95));
	dJointSetHingeParam(joint[1],dParamVel,(dReal) inputs[1]*2.55);
	dJointSetHingeParam(joint[1],dParamFMax,(dReal)2.55);
	// right foot
	dJointSetHingeParam(joint[2],dParamVel,(dReal)inputs[2]*3.2);
	dJointSetHingeParam(joint[2],dParamFMax,(dReal)3.7);
	// left foot
	dJointSetHingeParam(joint[3],dParamVel,(dReal)inputs[3]*3.2);
	dJointSetHingeParam(joint[3],dParamFMax,(dReal)3.7);	
	// right foot
	dJointSetHingeParam(joint[4],dParamVel,(dReal)inputs[4]*1.7);
	dJointSetHingeParam(joint[4],dParamFMax,(dReal)2.5);
	// left foot
	dJointSetHingeParam(joint[5],dParamVel,(dReal)inputs[5]*1.7);
	dJointSetHingeParam(joint[5],dParamFMax,(dReal)2.5);	
}

void Box::getInputs(float * inputs) const
{
	// position inputs (position & rotation of head)
/*	const dReal * vel = dBodyGetLinearVel(body[5]);
	const dReal * rot = dBodyGetAngularVel(body[5]);
	const dReal * vel2 = dBodyGetLinearVel(body[6]);
	const dReal * rot2 = dBodyGetAngularVel(body[6]); */
//const dReal * vel = dBodyGetLinearVel(body[0]);
	const dReal * angle = dBodyGetRotation(body[0]);
//	cout << angle[1] << endl; 	// angle of body is angle[1]
	// get right leg socket angles (3)
//	inputs[14] = (float)dJointGetAMotorAngle(lmotor[7],0);
	inputs[0] = (float)dJointGetHingeAngleRate(joint[0]);
	inputs[1] = (float)dJointGetHingeAngleRate(joint[1]);
	// get left leg socket angles (3)
	// get right foot angle
	inputs[2] = (float)dJointGetHingeAngleRate(joint[2]);
	// get left foot angle
	inputs[3] = (float)dJointGetHingeAngleRate(joint[3]);
	// get right foot angle
	inputs[4] = (float)dJointGetHingeAngleRate(joint[4]);
	// get left foot angle
	inputs[5] = (float)dJointGetHingeAngleRate(joint[5]);
/*	inputs[0] = vel[0];
	inputs[1] = vel[1];
	inputs[2] = vel[2];
	inputs[3] = vel2[0];
	inputs[4] = vel2[1];
	inputs[5] = vel2[2];  */
//	inputs[6] = vel[1];
	inputs[6] = angle[1];
//	inputs[2] = rot[0];

	if (r_foot_sens)
		inputs[7] = 1.0;
	else	
		inputs[7] = -1.0;
	if (l_foot_sens)
		inputs[8] = 1.0;
	else
		inputs[8] = -1.0;

}

void Box::footContact(dGeomID o1, dGeomID o2, dGeomID ground)
{
	if (((o1==geom[5]) && (o2==ground)) || ((o2==geom[5])&&(o1==ground)))
		r_foot_sens=true;
	if (((o1==geom[6]) && (o2==ground)) || ((o2==geom[6])&&(o1==ground)))
		l_foot_sens=true;
}

void Box::clearSensor(void)
{
  r_foot_sens=false;
  l_foot_sens=false;
}

const dReal *Box::getHeadPos(void) const
{
	return dBodyGetPosition(body[0]);
}

const dReal *Box::getBodyRot(void) const
{
	return dBodyGetRotation(body[0]);
}

const float Box::getFitness(void) const
{
	const dReal *pos = getHeadPos();
  float result = -pos[1]*9 - pow(fabs(getStartX()-pos[0]),4);
  // starting pos = 0
//  float dif = 2/(pow((-7.5 - pos[1]),2) + fabs(getStartX()-pos[0]) + 0.01);

//  float dif = pow(fabs(pos[1] + 7.5),2)+0.1;
//  float result = (24*dif) - (pow(fabs(getStartX()-pos[0]),2)*0.25);

	if (result > 0)
		return result;
	else
		return 0;		// y coordinate.
}

void Box::randomForce(void) const
{
    dReal f[3];
    f[0] = 22.5 * (-0.5+1.0/RAND_MAX * rand());
    f[1] = 42.5 * (-0.5+1.0/RAND_MAX * rand());
    f[2] = .5 * (-0.5+1.0/RAND_MAX * rand());
    dBodyAddForce(body[0], f[0], f[1], f[2]);
}
void Box::randomTorque(void) const
{
    dReal f[3];
    f[0] = 5.72 * (-0.5+1.0/RAND_MAX * rand());
    f[1] = 5.72 * (-0.5+1.0/RAND_MAX * rand());
    f[2] = 5.72 * (-0.5+1.0/RAND_MAX * rand());
	  dBodyAddTorque(body[0], f[0], f[1], f[2]);
}


void Box::bodyContact(dGeomID o1, dGeomID o2, dGeomID ground)
{
	// dummy function
}

bool Box::hasFallen(void) const
{
	return false;
}

void Box::shutoff(void)
{
	// dummy function
}

void Box::printStats(void) const
{
	const dReal *pos = dBodyGetPosition(body[0]);
	const dReal *rot = dBodyGetRotation(body[0]);

	cout << "\nPosition:";
	cout << "X: " << pos[0] << " ";
	cout << "Y: " << pos[1] << " ";
	cout << "Z: " << pos[2] << endl;
	cout << "Rotation:" << endl;
	cout << "XR: " << rot[0] << " ";
	cout << "YR: " << rot[1] << " ";
	cout << "ZR: " << rot[2] << endl;
	cout << "Feet:" << endl;
	cout << "Right: " << r_foot_sens << " ";
	cout << "Left: " << l_foot_sens << " " << endl;
//  cout << "Sx: " << getStartX() << "Sy: " << getStartY() << endl;
}

