/***************************************************************************
                          hexapod.h  -  description
                             -------------------
    begin                : Sun Mar 17 2002
    copyright            : (C) 2002 by Gene Ruebsamen
    email                : 
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef HEXAPOD_H
#define HEXAPOD_H
#include <ode/ode.h>
#include <drawstuff/drawstuff.h>
#include "creature.h"

class Hexapod : public Creature {
public:
	Hexapod(dWorldID,dSpaceID,dReal,dReal,dReal);		// constructor (world,space,x,y,z,in,hid,out,recur)
	~Hexapod();	// destructor
	void draw(bool=false) const;
	void getInputs(float *) const;
	const dReal * getHeadPos(void) const;
	const dReal * getBodyRot(void) const;
	const float getFitness(void) const;
	void setMuscles(const float * const);
	void footContact(dGeomID,dGeomID,dGeomID);   		// are the feet touching the ground
	void bodyContact(dGeomID,dGeomID,dGeomID);   		// are the body touching the ground
	void randomForce(void) const;
	void randomTorque(void) const;
	void shutoff(void);
	bool hasFallen(void) const;
	void printStats(void) const;
  void clearSensor(void);
private:
	void constructBody(dWorldID,dSpaceID,dReal,dReal,dReal);
	const int numBody;
	const int numGeom;
	const int numJoint;
	bool l_foot_sens;
	bool r_foot_sens;
	dBodyID *body;
	dGeomID *geom;
	dJointID *joint;
	dJointGroupID jGroup;
	dGeomID geomGroup;
	//  declarations
	dReal TORSO[3]; 	// length, width height
	dReal LEG[3];	
	dReal GAP;
};

#endif

