/***************************************************************************
                          human.cpp  -  description
                             -------------------
    begin                : Mon Feb 18 2002
    copyright            : (C) 2002 by Gene Ruebsamen
    email                : 
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "human.h"      // import constants for bodies

Human::Human(dWorldID world, dSpaceID space, dReal x, dReal y, dReal z) :
	Creature(world,space,x,y,z), numBody(12), numGeom(8), numJoint(15)
{
	// setup
	HEAD = 0.12;
	TORSO[0] = 0.4; TORSO[1] = 0.25; TORSO[2] = 0.60;
	ARM1[0] = 0.10; ARM1[1] = 0.10; ARM1[2] = 0.27;
	ARM2[0] = 0.10; ARM2[1] = 0.10; ARM2[2] = 0.25;
	LEG1[0] = 0.15; LEG1[1] = 0.15; LEG1[2] = 0.4;
	LEG2[0] = 0.15; LEG2[1] = 0.15; LEG2[2] = 0.37;
	FOOT[0] = 0.20; FOOT[1] = 0.35; FOOT[2] = 0.05;
	GAP = 0.01;

	l_foot_sens = false;					// init foot sensors
	r_foot_sens = false;
	body = new dBodyID[numBody];	// allocate space for Bodies
	geom = new dGeomID[numGeom];	// allocate space for Geometries	
	joint = new dJointID[numJoint];	// allocate space for Joints
	
	// create new Joint Group
	jGroup = dJointGroupCreate(0);
	geomGroup = dCreateGeomGroup(space);

	// constants (Mass)
	const dReal MHEAD=0.06;
	const dReal MTORSO=0.5;
	const dReal MUARM=0.2;
	const dReal MLARM=0.10;
	const dReal MULEG=0.35;
	const dReal MLLEG=0.3;
	const dReal MFOOT=0.05;

	dMass m;

	// create bodies
	for (int i=0;i<numBody;i++)
		body[i]=dBodyCreate(world);

	dBodySetPosition (body[0],x,y,z+TORSO[2]/2+HEAD);	// head
	dBodySetPosition (body[1],x,y,z);	// torso
	dBodySetPosition (body[2],TORSO[0]/2+ARM1[1]/2+GAP/4+x,y,z+(TORSO[2]/2-ARM1[2]/2));	//right upper arm
	dBodySetPosition (body[3],-TORSO[0]/2-ARM1[1]/2-GAP/4+x,y,z+(TORSO[2]/2-ARM1[2]/2));	//left upper arm
	dBodySetPosition (body[4],TORSO[0]/2+ARM2[1]/2+GAP/4+x,y,z+(TORSO[2]/2-ARM1[2]/2-ARM1[2]-GAP));	// right lower arm
	dBodySetPosition (body[5],-TORSO[0]/2-ARM2[1]/2-GAP/4+x,y,z+(TORSO[2]/2-ARM1[2]/2-ARM1[2]-GAP));	// left lower arm
	dBodySetPosition (body[6],TORSO[0]/3+x,0+y,z-(TORSO[2]/2+LEG1[2]/2+GAP));	//right upper leg
	dBodySetPosition (body[7],-TORSO[0]/3+x,0+y,z-(TORSO[2]/2+LEG1[2]/2+GAP));	// left upper leg
	dBodySetPosition (body[8],TORSO[0]/3+x,0+y,z-(TORSO[2]/2+LEG1[2]/2+LEG1[2]));	// right lower leg
	dBodySetPosition (body[9],-TORSO[0]/3+x,0+y,z-(TORSO[2]/2+LEG1[2]/2+LEG1[2]));	//left lower leg
	dBodySetPosition (body[10],TORSO[0]/3+GAP/2+x,-FOOT[0]/4+y,z-TORSO[2]/2-LEG1[2]-LEG2[2]-8*GAP);	//right foot
	dBodySetPosition (body[11],-TORSO[0]/3-GAP/2+x,-FOOT[0]/4+y,z-TORSO[2]/2-LEG1[2]-LEG2[2]-8*GAP);	//left foot

	// head
	dMassSetSphere(&m,1,HEAD);
	dMassAdjust(&m,MHEAD);
	dBodySetMass(body[0], &m);
	geom[0] = dCreateSphere(space,HEAD);
	dGeomSetBody(geom[0],body[0]);

	// torso
	dMassSetBox (&m,1,TORSO[0],TORSO[1],TORSO[2]);	// torso
	dMassAdjust (&m,MTORSO);	// torso weight
	dBodySetMass (body[1],&m);
	geom[1] = dCreateBox (0,TORSO[0],TORSO[1],TORSO[2]);
	dGeomSetBody (geom[1],body[1]);

	// upper arms
	dMassSetBox (&m,1,ARM1[0],ARM1[1],ARM1[2]);	// upper arms
	dMassAdjust (&m,MUARM);	// upper arm weight
	for (int i=2;i<=3;i++)
	{
    	dBodySetMass (body[i],&m);
//    	geom[i] = dCreateBox (space,ARM1[0],ARM1[1],ARM1[2]);
 //   	dGeomSetBody (geom[i],body[i]);
  }
	// lower arms
	dMassSetBox (&m,1,ARM2[0],ARM2[1],ARM2[2]);	// lower arms
	dMassAdjust (&m,MLARM);	// lower arm weight
  geom[2] = dCreateBox (0,ARM2[0],ARM2[1],ARM2[2]);
  dGeomSetBody (geom[2],body[4]);
  geom[3] = dCreateBox (0,ARM2[0],ARM2[1],ARM2[2]);
  dGeomSetBody (geom[3],body[5]);
	for (int i=4;i<=5;i++)
		dBodySetMass (body[i],&m);
	// upper legs
	dMassSetBox (&m,1,LEG1[0],LEG1[1],LEG1[2]);	// legs
	dMassAdjust (&m,MULEG);	// upper leg weight
	for (int i=6;i<=7;i++)
		dBodySetMass (body[i],&m);
	// lower legs
	dMassSetBox (&m,1,LEG2[0],LEG2[1],LEG2[2]);	// legs
  dMassAdjust (&m,MLLEG);	// lower leg weight
		geom[4] = dCreateBox (0,LEG2[0],LEG2[1],LEG2[2]);
		dGeomSetBody (geom[4],body[8]);
		geom[5] = dCreateBox (0,LEG2[0],LEG2[1],LEG2[2]);
		dGeomSetBody (geom[5],body[9]);
	for (int i=8;i<=9;i++)
		dBodySetMass (body[i],&m);
	// feet
	dMassSetBox (&m,1,FOOT[0],FOOT[1],FOOT[2]); 	// feet
	dMassAdjust (&m,MFOOT);	// feet weight
		geom[6] = dCreateBox (0,FOOT[0],FOOT[1],FOOT[2]);  // 0
		dGeomSetBody (geom[6],body[10]);
		geom[7] = dCreateBox (0,FOOT[0],FOOT[1],FOOT[2]);  // 0
		dGeomSetBody (geom[7],body[11]);
	for (int i=10;i<=11;i++)
		dBodySetMass (body[i],&m);

	// add Geometries to geomGroup
	for (int i=0;i<numGeom;i++)
		dGeomGroupAdd(geomGroup,geom[i]);


	/********************************** CREATE JOINTS **********************************/
	// head joint
	joint[0] = dJointCreateFixed(world,jGroup);		// add to joint group jGroup
	dJointAttach(joint[0],body[1],body[0]);
	dJointSetFixed(joint[0]);
	// right shoulder
	joint[1] = dJointCreateBall(world,0);
	dJointAttach(joint[1],body[1],body[2]);
	dJointSetBallAnchor(joint[1],TORSO[0]/2+ARM1[1]/2+GAP/8+x,y,z+(TORSO[2]/2));
	joint[11] = dJointCreateAMotor(world,0);
	dJointAttach(joint[11],body[1],body[2]);
	// left shoulder
	joint[2] = dJointCreateBall(world,0);
	dJointAttach(joint[2],body[1],body[3]);
	dJointSetBallAnchor(joint[2],-TORSO[0]/2-ARM1[1]/2-GAP/8+x,y,z+(TORSO[2]/2));
	joint[12] = dJointCreateAMotor(world,0);
	dJointAttach(joint[12],body[1],body[3]);
	dJointSetAMotorParam(joint[11],dParamLoStop3,0);	// left
	dJointSetAMotorParam(joint[11],dParamHiStop3,2.0);  // right
	dJointSetAMotorParam(joint[12],dParamLoStop3,-2.0);	// left
	dJointSetAMotorParam(joint[12],dParamHiStop3,0);  // right
	for (int i=11;i<=12;i++)
	{          // ARM MOTORS
		dJointSetAMotorMode(joint[i],dAMotorEuler);
		dJointSetAMotorAxis(joint[i],0,1,0,0,1);
		dJointSetAMotorAxis(joint[i],2,2,0,1,0);
		dJointSetAMotorParam(joint[i],dParamLoStop,-0.2);  // twist
		dJointSetAMotorParam(joint[i],dParamHiStop,0.2);
		dJointSetAMotorParam(joint[i],dParamVel,0);
		dJointSetAMotorParam(joint[i],dParamFMax,0.05); 		// joint disabled
		dJointSetAMotorParam(joint[i],dParamLoStop2,-0.2);  // back
		dJointSetAMotorParam(joint[i],dParamHiStop2,1.5);   // front
		dJointSetAMotorParam(joint[i],dParamVel2,0);
		dJointSetAMotorParam(joint[i],dParamFMax2,0.05);
		dJointSetAMotorParam(joint[i],dParamVel3,0);
		dJointSetAMotorParam(joint[i],dParamFMax3,2);
		dJointSetAMotorParam(joint[i],dParamStopERP,0.2);
		dJointSetAMotorParam(joint[i],dParamStopCFM,0.03);
		dJointSetAMotorParam(joint[i],dParamStopERP2,0.2);
		dJointSetAMotorParam(joint[i],dParamStopCFM2,0.03);
		dJointSetAMotorParam(joint[i],dParamStopERP3,0.2);
		dJointSetAMotorParam(joint[i],dParamStopCFM3,0.03);
	}
	// right elbow
	joint[3] = dJointCreateHinge(world,0);
	dJointAttach(joint[3],body[2],body[4]);
	dJointSetHingeAnchor(joint[3],TORSO[0]/2+ARM2[1]/2+GAP/8+x,y,z+(TORSO[2]/2-ARM1[2]-GAP/2));
	dJointSetHingeAxis(joint[3],1,0,0);
	dJointSetHingeParam(joint[3],dParamLoStop,0);
	dJointSetHingeParam(joint[3],dParamHiStop,1.90);
	dJointSetHingeParam(joint[3],dParamVel,0);
	dJointSetHingeParam(joint[3],dParamFMax,10);
	dJointSetHingeParam(joint[3],dParamStopERP,0.2);
	dJointSetHingeParam(joint[3],dParamStopCFM,0.03);
	// left elbow
	joint[4] = dJointCreateHinge(world,0);
	dJointAttach(joint[4],body[3],body[5]);
	dJointSetHingeAnchor(joint[4],-TORSO[0]/2-ARM2[1]/2-GAP/8+x,y,z+(TORSO[2]/2-ARM1[2]-GAP/2));
	dJointSetHingeAxis(joint[4],1,0,0);
	dJointSetHingeParam(joint[4],dParamLoStop,0);
	dJointSetHingeParam(joint[4],dParamHiStop,1.90);
	dJointSetHingeParam(joint[4],dParamVel,0);
	dJointSetHingeParam(joint[4],dParamFMax,10);
	dJointSetHingeParam(joint[4],dParamStopERP,0.2);
	dJointSetHingeParam(joint[4],dParamStopCFM,0.03);
	// right Leg Socket
	joint[5] = dJointCreateBall(world,jGroup);
	dJointAttach(joint[5],body[1],body[6]);
	dJointSetBallAnchor(joint[5],TORSO[0]/3+x,0+y,z-(TORSO[2]/2)-GAP/2);
	joint[13] = dJointCreateAMotor(world,jGroup);
	dJointAttach(joint[13],body[1],body[6]);

	// left Leg Socket
	joint[6] = dJointCreateBall(world,jGroup);
	dJointAttach(joint[6],body[1],body[7]);
	dJointSetBallAnchor(joint[6],-TORSO[0]/3+x,0+y,z-(TORSO[2]/2)-GAP/2);
	joint[14] = dJointCreateAMotor(world,jGroup);
	dJointAttach(joint[14],body[1],body[7]);
	dJointSetAMotorParam(joint[13],dParamLoStop3,-0.2); //left
	dJointSetAMotorParam(joint[13],dParamHiStop3,0.6);  // right
	dJointSetAMotorParam(joint[14],dParamLoStop3,-0.6); //left
	dJointSetAMotorParam(joint[14],dParamHiStop3,0.2);  // right

	for (int i=13;i<=14;i++)
	{        // LEG MOTORS
		dJointSetAMotorMode(joint[i],dAMotorEuler);
		dJointSetAMotorAxis(joint[i],0,1,0,0,1);
		dJointSetAMotorAxis(joint[i],2,2,0,1,0);
		dJointSetAMotorParam(joint[i],dParamLoStop,-0.2);  // twist
		dJointSetAMotorParam(joint[i],dParamHiStop,0.2);
		dJointSetAMotorParam(joint[i],dParamVel, 0);
		dJointSetAMotorParam(joint[i],dParamFMax,25);
		dJointSetAMotorParam(joint[i],dParamLoStop2,-0.9);  //back
		dJointSetAMotorParam(joint[i],dParamHiStop2,1.5);	// front
		dJointSetAMotorParam(joint[i],dParamVel2,0);
		dJointSetAMotorParam(joint[i],dParamFMax2,0.8);
		dJointSetAMotorParam(joint[i],dParamVel3,0);
		dJointSetAMotorParam(joint[i],dParamFMax3,0.8);
		dJointSetAMotorParam(joint[i],dParamStopERP,0.2);
		dJointSetAMotorParam(joint[i],dParamStopCFM,0.03);
		dJointSetAMotorParam(joint[i],dParamStopERP2,0.2);
		dJointSetAMotorParam(joint[i],dParamStopCFM2,0.03);
		dJointSetAMotorParam(joint[i],dParamStopERP3,0.2);
		dJointSetAMotorParam(joint[i],dParamStopCFM3,0.03);
	}
	// right knee
	joint[7] = dJointCreateHinge(world,jGroup);
	dJointAttach(joint[7],body[6],body[8]);
	dJointSetHingeAnchor(joint[7],TORSO[0]/3+x,y,z-(TORSO[2]/2+LEG1[2]));
	dJointSetHingeAxis(joint[7],1,0,0);
	dJointSetHingeParam(joint[7],dParamLoStop,-2.2);
	dJointSetHingeParam(joint[7],dParamHiStop,0);
	dJointSetHingeParam(joint[7],dParamVel,0);
	dJointSetHingeParam(joint[7],dParamFMax,0.5);
	dJointSetHingeParam(joint[7],dParamStopERP,0.2);
	dJointSetHingeParam(joint[7],dParamStopCFM,0.03);
	// left knee
	joint[8] = dJointCreateHinge(world,jGroup);
	dJointAttach(joint[8],body[7],body[9]);
	dJointSetHingeAnchor(joint[8],-TORSO[0]/3+x,y,z-(TORSO[2]/2+LEG1[2]));
	dJointSetHingeAxis(joint[8],1,0,0);
	dJointSetHingeParam(joint[8],dParamLoStop,-2.2);
	dJointSetHingeParam(joint[8],dParamHiStop,0);
	dJointSetHingeParam(joint[8],dParamVel,0);
	dJointSetHingeParam(joint[8],dParamFMax,0.5);
	dJointSetHingeParam(joint[8],dParamStopERP,0.2);
	dJointSetHingeParam(joint[8],dParamStopCFM,0.03);
	// right foot
	joint[9] = dJointCreateHinge(world,jGroup);
	dJointAttach(joint[9],body[8],body[10]);
	dJointSetHingeAnchor(joint[9],TORSO[0]/3+GAP/2+x,0.06+y,z-TORSO[2]/2-LEG1[2]-LEG2[2]-(8*GAP/2));
	dJointSetHingeAxis(joint[9],1,0,0);
	dJointSetHingeParam(joint[9],dParamLoStop,-0.8);
	dJointSetHingeParam(joint[9],dParamHiStop,0.3);
	dJointSetHingeParam(joint[9],dParamStopERP,0.2);
	dJointSetHingeParam(joint[9],dParamStopCFM,0.03);
	dJointSetHingeParam(joint[9],dParamBounce,0.3);
	dJointSetHingeParam(joint[9],dParamFudgeFactor,0.5);
	dJointSetHingeParam(joint[9],dParamVel,0);
	dJointSetHingeParam(joint[9],dParamFMax,0.5);
	// left foot
	joint[10] = dJointCreateHinge(world,jGroup);
	dJointAttach(joint[10],body[9],body[11]);
	dJointSetHingeAnchor(joint[10],-TORSO[0]/3-GAP/2+x,0.06+y,z-TORSO[2]/2-LEG1[2]-LEG2[2]-(8*GAP/2));
	dJointSetHingeAxis(joint[10],1,0,0);
	dJointSetHingeParam(joint[10],dParamLoStop,-0.8);
	dJointSetHingeParam(joint[10],dParamHiStop,0.3);
	dJointSetHingeParam(joint[10],dParamStopERP,0.2);
	dJointSetHingeParam(joint[10],dParamStopCFM,0.03);
	dJointSetHingeParam(joint[10],dParamBounce,0.3);
	dJointSetHingeParam(joint[10],dParamFudgeFactor,0.5);
	dJointSetHingeParam(joint[10],dParamVel,0);
	dJointSetHingeParam(joint[10],dParamFMax,0.5);
}

Human::~Human()
{
	// destroy joint group
	dJointGroupDestroy(jGroup);
	// destroy Geometries
	for (int i=0;i<numGeom;i++) {
		dGeomGroupRemove(geomGroup, geom[i]);
		dGeomDestroy(geom[i]);
	}
	// destroy bodies
	for (int i=0;i<numBody;i++)
		dBodyDestroy(body[i]);

	// deallocate space
	delete [] body;
	delete [] geom;
	delete [] joint;
}

void Human::draw(bool selected) const
{
	dsSetColor (0,1,1);
	dsSetTexture (DS_WOOD); // DS_WOOD
	// head
	dsDrawSphere(dBodyGetPosition(body[0]),dBodyGetRotation(body[0]),HEAD);
	if (selected)
		dsSetColor(1,0,0);
  else
		dsSetColor (1,0,1);
	// torso
	dsDrawBox(dBodyGetPosition(body[1]),dBodyGetRotation(body[1]),TORSO);
	// arms
	dsDrawBox(dBodyGetPosition(body[2]),dBodyGetRotation(body[2]),ARM1);
	dsDrawBox(dBodyGetPosition(body[3]),dBodyGetRotation(body[3]),ARM1);
	dsDrawBox(dBodyGetPosition(body[4]),dBodyGetRotation(body[4]),ARM2);
	dsDrawBox(dBodyGetPosition(body[5]),dBodyGetRotation(body[5]),ARM2);
	dsSetColor (1,1,0);
	// legs
	dsDrawBox(dBodyGetPosition(body[6]),dBodyGetRotation(body[6]),LEG1);
	dsDrawBox(dBodyGetPosition(body[7]),dBodyGetRotation(body[7]),LEG1);
	dsDrawBox(dBodyGetPosition(body[8]),dBodyGetRotation(body[8]),LEG2);
	dsDrawBox(dBodyGetPosition(body[9]),dBodyGetRotation(body[9]),LEG2);
	dsSetColor (0.5,0,1);
	// feet
	dsDrawBox(dBodyGetPosition(body[10]),dBodyGetRotation(body[10]),FOOT);
	dsDrawBox(dBodyGetPosition(body[11]),dBodyGetRotation(body[11]),FOOT);
}

void Human::setMuscles(const float * const inputs)
{
	// right ARM Socket
	dJointSetAMotorParam(joint[11],dParamVel,(dReal) inputs[0]);
	dJointSetAMotorParam(joint[11],dParamFMax, (dReal)1);
	dJointSetAMotorParam(joint[11],dParamVel2,(dReal) inputs[1]);
	dJointSetAMotorParam(joint[11],dParamFMax2,(dReal)1);
	dJointSetAMotorParam(joint[11],dParamVel3,(dReal) inputs[2]);
	dJointSetAMotorParam(joint[11],dParamFMax3,(dReal)1);
	// left ARM socket
	dJointSetAMotorParam(joint[12],dParamVel,(dReal) inputs[3]);
	dJointSetAMotorParam(joint[12],dParamFMax,(dReal)1);
	dJointSetAMotorParam(joint[12],dParamVel2,(dReal) inputs[4]);
	dJointSetAMotorParam(joint[12],dParamFMax2,(dReal)1);
	dJointSetAMotorParam(joint[12],dParamVel3,(dReal) inputs[5]);
	dJointSetAMotorParam(joint[12],dParamFMax3,(dReal)1);
	// right elbow
	dJointSetHingeParam(joint[3],dParamVel,(dReal)inputs[6]*1.95);
	dJointSetHingeParam(joint[3],dParamFMax,(dReal)1);
	// left elbow
	dJointSetHingeParam(joint[4],dParamVel,(dReal)inputs[7]*1.95);
	dJointSetHingeParam(joint[4],dParamFMax,(dReal)1);

	// right Leg Socket
	dJointSetAMotorParam(joint[13],dParamVel,(dReal) inputs[8]*1.25);   // twist
	dJointSetAMotorParam(joint[13],dParamFMax, (dReal)1.55);
	dJointSetAMotorParam(joint[13],dParamVel2,(dReal) inputs[9]*1.95);
	dJointSetAMotorParam(joint[13],dParamFMax2,(dReal)4.55);
	dJointSetAMotorParam(joint[13],dParamVel3,(dReal) inputs[10]);       // left/right
	dJointSetAMotorParam(joint[13],dParamFMax3,(dReal)2.95);
	// left leg socket
	dJointSetAMotorParam(joint[14],dParamVel,(dReal) inputs[11]*1.25);   // twist
	dJointSetAMotorParam(joint[14],dParamFMax,(dReal)1.25);
	dJointSetAMotorParam(joint[14],dParamVel2,(dReal) inputs[12]*1.95);
	dJointSetAMotorParam(joint[14],dParamFMax2,(dReal)4.55);
	dJointSetAMotorParam(joint[14],dParamVel3,(dReal) inputs[13]);				// left/right
	dJointSetAMotorParam(joint[14],dParamFMax3,(dReal)2.95);
	// right knee
	dJointSetHingeParam(joint[7],dParamVel,(dReal)inputs[14]*1.95);
	dJointSetHingeParam(joint[7],dParamFMax,(dReal)3.95);
	// left knee
	dJointSetHingeParam(joint[8],dParamVel,(dReal)inputs[15]*1.95);
	dJointSetHingeParam(joint[8],dParamFMax,(dReal)3.95);
	// right foot
	dJointSetHingeParam(joint[9],dParamVel,(dReal)inputs[16]);
	dJointSetHingeParam(joint[9],dParamFMax,(dReal)1.7);
	// left foot
	dJointSetHingeParam(joint[10],dParamVel,(dReal)inputs[17]);
	dJointSetHingeParam(joint[10],dParamFMax,(dReal)1.7);	
}

void Human::getInputs(float * inputs) const
{
	// position inputs (position & rotation of head)
	const dReal * pos = dBodyGetAngularVel(body[0]);
	const dReal * rot = dBodyGetRotation(body[0]);
	inputs[0] = (float)pos[0];		// z (height) of head
	inputs[1] = (float)pos[1];		// z (height) of head
	inputs[2] = (float)pos[2];		// z (height) of head
	inputs[3] = (float)rot[0];
	inputs[4] = (float)rot[1];
	inputs[5] = (float)rot[2];
	// get right arm socket angles (3)
	inputs[6] = (float)dJointGetAMotorAngle(joint[11],0);
	inputs[7] = (float)dJointGetAMotorAngle(joint[11],1);
	inputs[8] = (float)dJointGetAMotorAngle(joint[11],2);
	// get left arm socket angles (3)
	inputs[9] = (float)dJointGetAMotorAngle(joint[12],0);
	inputs[10] = (float)dJointGetAMotorAngle(joint[12],1);
	inputs[11] = (float)dJointGetAMotorAngle(joint[12],2);
	// get right elbow angle
	inputs[12] = (float)dJointGetHingeAngleRate(joint[3]);
	// get left elbow angle
	inputs[13] = (float)dJointGetHingeAngleRate(joint[4]);
	// get right leg socket angles (3)
	inputs[14] = (float)dJointGetAMotorAngle(joint[13],0);
	inputs[15] = (float)dJointGetAMotorAngle(joint[13],1);
	inputs[16] = (float)dJointGetAMotorAngle(joint[13],2);
	// get left leg socket angles (3)
	inputs[17] = (float)dJointGetAMotorAngle(joint[14],0);
	inputs[18] = (float)dJointGetAMotorAngle(joint[14],1);
	inputs[19] = (float)dJointGetAMotorAngle(joint[14],2);
	// get right knee angle
	inputs[20] = (float)dJointGetHingeAngleRate(joint[7]);
	// get left knee angle
	inputs[21] = (float)dJointGetHingeAngleRate(joint[8]);
	// get right foot angle
	inputs[22] = (float)dJointGetHingeAngleRate(joint[9]);
	// get left foot angle
	inputs[23] = (float)dJointGetHingeAngleRate(joint[10]);
	if (r_foot_sens) 	inputs[24] = 1.0;
	else	inputs[24] = -1.0;
	if (l_foot_sens)	inputs[25] = 1.0;
	else	inputs[25] = -1.0;
}

void Human::shutoff(void)
{
	// right Leg Socket
	dJointSetHingeParam(joint[3],dParamVel,(dReal)0);
	dJointSetHingeParam(joint[3],dParamFMax, (dReal)0);
	// left leg socket
	dJointSetHingeParam(joint[4],dParamVel,(dReal)0);
	dJointSetHingeParam(joint[4],dParamFMax,(dReal)0);
	// right foot
	dJointSetHingeParam(joint[7],dParamVel,(dReal)0);
	dJointSetHingeParam(joint[7],dParamFMax,(dReal)0);
	// left foot
	dJointSetHingeParam(joint[8],dParamVel,(dReal)0);
	dJointSetHingeParam(joint[8],dParamFMax,(dReal)0);	
	dJointSetHingeParam(joint[9],dParamVel,(dReal)0);
	dJointSetHingeParam(joint[9],dParamFMax,(dReal)0);	
	dJointSetHingeParam(joint[10],dParamVel,(dReal)0);
	dJointSetHingeParam(joint[10],dParamFMax,(dReal)0);	
	dJointSetHingeParam(joint[11],dParamVel,(dReal)0);
	dJointSetHingeParam(joint[11],dParamFMax,(dReal)0);	
	dJointSetHingeParam(joint[11],dParamVel2,(dReal)0);
	dJointSetHingeParam(joint[11],dParamFMax2,(dReal)0);	
	dJointSetHingeParam(joint[11],dParamVel3,(dReal)0);
	dJointSetHingeParam(joint[11],dParamFMax3,(dReal)0);	
	dJointSetHingeParam(joint[12],dParamVel,(dReal)0);
	dJointSetHingeParam(joint[12],dParamFMax,(dReal)0);	
	dJointSetHingeParam(joint[12],dParamVel2,(dReal)0);
	dJointSetHingeParam(joint[12],dParamFMax2,(dReal)0);	
	dJointSetHingeParam(joint[12],dParamVel3,(dReal)0);
	dJointSetHingeParam(joint[12],dParamFMax3,(dReal)0);	
	dJointSetHingeParam(joint[13],dParamVel,(dReal)0);
	dJointSetHingeParam(joint[13],dParamFMax,(dReal)0);	
	dJointSetHingeParam(joint[13],dParamVel2,(dReal)0);
	dJointSetHingeParam(joint[13],dParamFMax2,(dReal)0);	
	dJointSetHingeParam(joint[13],dParamVel3,(dReal)0);
	dJointSetHingeParam(joint[13],dParamFMax3,(dReal)0);	
	dJointSetHingeParam(joint[14],dParamVel,(dReal)0);
	dJointSetHingeParam(joint[14],dParamFMax,(dReal)0);	
	dJointSetHingeParam(joint[14],dParamVel2,(dReal)0);
	dJointSetHingeParam(joint[14],dParamFMax2,(dReal)0);	
	dJointSetHingeParam(joint[14],dParamVel3,(dReal)0);
	dJointSetHingeParam(joint[14],dParamFMax3,(dReal)0);	
}

float Human::getFitness(void) const
{
	const dReal * pos = getHeadPos();
	float fitness;

	// fitness function
	if (fabs(pos[1]) > 10 || (fabs(pos[2]) > 5) || (fabs(pos[0]-getStartX()) > 5))
		fitness = 0;
	else if (hasFallen())
		fitness = -pos[1]+(pos[2]*8);  	// if fallen down
	else
		fitness = -(pos[1]*2)+(pos[2]*8); // if still standing

	return fitness;
}

void Human::footContact(dGeomID o1, dGeomID o2, dGeomID ground)
{
	if (((o1==geom[10]) && (o2==ground)) || ((o2==geom[10])&&(o1==ground)))
		r_foot_sens=true;
	else
		r_foot_sens=false;
	if (((o1==geom[11]) && (o2==ground)) || ((o2==geom[11])&&(o1==ground)))
		l_foot_sens=true;
	else
		l_foot_sens=false;
}

void Human::bodyContact(dGeomID o1, dGeomID o2, dGeomID ground)
{
	if (((o1==geom[1]) && (o2==ground)) || ((o2==geom[1])&&(o1==ground)))
	{
		body_sens=true;
		fallen=true;
	}
	else
		body_sens=false;
}

const dReal *Human::getHeadPos(void) const
{
	return dBodyGetPosition(body[0]);
}

const dReal *Human::getRFootPos(void) const
{
	return dBodyGetPosition(body[10]);
}

const dReal *Human::getLFootPos(void) const
{
	return dBodyGetPosition(body[11]);
}

bool Human::hasFallen(void) const
{
	return fallen;
}

void Human::randomForce(void) const
{
    dReal f[3];
    f[0] = 22.5 * (-0.5+1.0/RAND_MAX * rand());
    f[1] = 42.5 * (-0.5+1.0/RAND_MAX * rand());
    f[2] = .5 * (-0.5+1.0/RAND_MAX * rand());
    dBodyAddForce(body[0], f[0], f[1], f[2]);
}
void Human::randomTorque(void) const
{
    dReal f[3];
    f[0] = 5.72 * (-0.5+1.0/RAND_MAX * rand());
    f[1] = 5.72 * (-0.5+1.0/RAND_MAX * rand());
    f[2] = 5.72 * (-0.5+1.0/RAND_MAX * rand());
	  dBodyAddTorque(body[0], f[0], f[1], f[2]);
}
